//===============================================================================
// Copyright 2014 Alachisoft.  All rights reserved.
//============================================================================

using System;
using System.Threading;
using System.Collections;
using System.Data;
using System.IO;

using Alachisoft.NCache.Web.Caching;
using Alachisoft.NCache.Runtime;

namespace CompactSerializationSample
{
	/// <summary>
	///Here is what this program should do:
	///
	///1. Create three classes "NormalObject", "BigObject", and "BiggerObject".
	///They are 4k, 10k, 50k in size respectively. Use C# class for each of the
	///above and within the class you can put a byte-array to take up the desired
	///size if you want but have at least two data members in the class (e.g.
	///String _id being the first data member).
	///
	///2. Write a command line program that goes into a loop (10,000 times) and for
	///each iteration, it uses the iteration number as the key and then randomly
	///picks one of the objects to add to the cache if GET on this key returns a
	///NULL.
	///
	///3. At the end of this loop, write another loop that goes through a part of
	///the first loop (e.g. 1/10 of the first loop) and does a GET on item (a
	///subset of the first loop). It should find all of these items in the Cache.
	///
	///4. Print out to a log-file (or a standard-output) for each iteration in the
	///loop. Print the following information on ONE LINE:
	///- print key
	///- found or not found
	///- Time it took to do a GET
	///- Time it took to do an INSERT (if insert is done)
	///- a sample line should look like this:
	///
	///1005  Found  GET 20ms
	///1006  Not    GET 50ms  INSERT  200ms
	///
	///5. Calculate the same statistics that you have done in "Program" (best,
	///average, worst, and also different bracket of worst) and print out these
	///statistics at the end.
	///
	///6. Put a 10 second wait after doing Initialize on the cache so when we run
	///it on two machines, the cluster is formed before data is actually added to
	///the cache.
	/// </summary>
	class Program
	{
       /// <summary> Number of iterations </summary>
		private static int RunCount = 100;
		/// <summary> Number of checkpoints </summary>
		private static int Checkpoint = Program.RunCount / 10;
		/// <summary> Part of the main loop for GET operations</summary>
		private const int Part = 3; // value should be between 1 to 10

		private static TimeStat 
			AddStats = new TimeStat(60, 120),
			GetStats = new TimeStat(1, 10);

		private static FileStream FlStream = new FileStream("cs-log.txt", FileMode.Create);
		private static StreamWriter StWriter = new StreamWriter(FlStream);
		private static TextWriter TxWriter = Console.Out;

		private static int userChoice = 1;

        /// <summary> Cache instance </summary>
        public static Cache _cache;

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			Console.SetOut(Program.TxWriter);
			Console.Write("Please specify a cache-id to benchmark: ");
			string CacheId = Console.ReadLine();

			try
			{
				Console.SetOut(Program.StWriter);
				Console.WriteLine("Initializing " + CacheId + "...\r\n");

				Console.SetOut(Program.TxWriter);
				Console.WriteLine("Initializing " + CacheId + "...");
				_cache = NCache.InitializeCache(CacheId);
                _cache.ExceptionsEnabled = true;
                Console.WriteLine("Cache initialized successfully...");
				
				Console.WriteLine("1. Small objects  (512 bytes)");
				Console.WriteLine("2. Normal objects (  4 K)");
				Console.WriteLine("3. Big objects    ( 10 K)");
				Console.WriteLine("4. Bigger objects ( 50 K)");
				Console.WriteLine("5. Sample objects ( >5 K)");
                Console.WriteLine("6. Mix all objects randomly");
				Console.Write("? > ");

				string userChoice = Console.ReadLine();
				Program.userChoice = Convert.ToInt32(userChoice);

				Console.Write("\nNumber of objects to add (default: " + Program.RunCount + "): ");
				userChoice = Console.ReadLine();
				if (!userChoice.Trim().Equals(""))
					Program.RunCount = Convert.ToInt32(userChoice);				
				
			}
			catch(Exception e)
			{
				Console.SetOut(Program.StWriter);
                Console.WriteLine("Error occured while trying to initialize Cache named: [" + CacheId + "]\nExceptin: " + e.Message);
				Console.SetOut(Program.TxWriter);
                Console.WriteLine("Error occured while trying to initialize Cache named: [" + CacheId + "]\nExceptin: " + e.Message);
                Console.WriteLine("Press any key to exit program...");
                Console.ReadKey();
				return;
			}
			
			Console.SetOut(Program.StWriter);

			Console.WriteLine("\r\nExpected results are:");
			Console.WriteLine("Operation    Best(ms)    Avg(ms)    Worst(ms)");
			Console.WriteLine("---------    --------    -------    ---------");
			Console.WriteLine("Add()        <= {0,-5}   <= {1,-5}  > {1,-5}", AddStats.ExpectedBest, AddStats.ExpectedWorst);
			Console.WriteLine("Get()        <= {0,-5}   <= {1,-5}  > {1,-5}", GetStats.ExpectedBest, GetStats.ExpectedWorst);
			
			Console.WriteLine("\r\nCompact Serialization Sample Started...\r\n");

			RunSimulation();
			DisplayStatistics();

			StWriter.Flush();

			Console.SetOut(Program.TxWriter);
			Console.WriteLine("\nPlease see performance statistics in cs-log.txt");
			Console.WriteLine("Press any key to exit program...");
			Console.ReadKey();
            _cache.Dispose();
			Program.Dispose();
		}

		/// <summary>
		/// performs the selected operation
		/// </summary>
		private static void RunSimulation()
		{
			try 
			{
				long objSize = 0;
				Console.WriteLine("Adding " + RunCount + " Items...");

				for(int i=0; i < RunCount; i++)
				{
					PerformOperation(i,ref objSize);
				}

				Console.WriteLine("\r\n-------------------------------------------------------------------------\r\n");
				Console.WriteLine("\r\nPerforming operations in parts...\r\n");

				for(int i = (Program.RunCount / 10) * Part ; i < (Program.RunCount / 10) * (Part + 1) ; i++)
				{
					PerformOperation(i, ref objSize);
				}
			}
			catch (Exception ex) 
			{
				Console.WriteLine("ERROR: " + ex.Message);
			}
		}

		/// <summary>
		/// Performs the get operation, if unsuccessful, performs an add operation, and log the time taken 
		/// to complete operation
		/// </summary>
		private static void PerformOperation(int i, ref long objSize)
		{
			Console.Write(i.ToString());
			object obj = Get(i.ToString());

			if(obj==null)
			{						
				Console.Write(" not found. Get operation took:" + GetStats.Current.ToString() + " ms\t");

				Console.Write("Inserting "+i);
				objSize = Add(i.ToString());
				Console.Write(", operation took:" + AddStats.Current.ToString() + " ms\t");
			}
			else
			{
				Console.Write(" found. Get operation took:" + GetStats.Current.ToString() + " ms\t");
			}

			if (objSize == 512)
				Console.WriteLine("Object size: 512 bytes");
			else 
				Console.WriteLine("Object size:" + (objSize / 1000) + " KB");
		}

		/// <summary>
		/// Adds an object to the cache if it is not already present.
		/// </summary>
		private static long Add(string key)
		{
			long objSize = 0;
			if (!_cache.Contains(key))
			{
				try
				{
					int operation = userChoice;
					if (Program.userChoice == 6) 
					{
						Random rand = new Random();
						operation = (int)rand.Next(1,5);
					}

					object Obj;
					switch(operation)
					{
						case 1:
							Obj = new SmallObject();
							objSize = 512;
							break;
						case 2:
							Obj = new NormalObject();
							objSize = 4000;
							break;
						case 3:
							Obj = new BigObject();
							objSize = 10000;
							break;
						case 4:
							Obj = new BiggerObject();
							objSize = 50000;
							break;
						default:
							Obj = new SampleObject();
							objSize = 50000;
							break;
					}

					AddStats.BeginSample();
					_cache.Add(
						key,
						Obj,
						null, 
						Cache.NoAbsoluteExpiration, 
						Cache.NoSlidingExpiration, 
						CacheItemPriority.Default
                        );
					AddStats.EndSample();
				}
				catch(Exception ex)
				{
					Console.WriteLine("ERROR: " + ex.Message);
				}
			}
			return objSize;
		}

		/// <summary>
		/// gets an object from the cache if present.
		/// </summary>
		private static Object Get(string key)
		{
			try
			{
				GetStats.BeginSample();
                Object obj = new object(); //NCache.Cache.Get(key);
                obj = _cache.Get(key);
				GetStats.EndSample();
				
				return obj;
			}
			catch(Exception)
			{
				return null;
			}
		}

		static private void DisplayStatistics()
		{
			Console.WriteLine("\r\nStatistics at: {0}", DateTime.Now.ToString());			
			Console.WriteLine("Operation    # Runs     Best(ms)    Avg(ms)    Worst(ms)");
			Console.WriteLine("---------    ------     --------    -------    ---------");
			Console.WriteLine("Add()        {0,-6}     {1,-8}    {2,-8:F2}   {3,-8}", AddStats.Runs, AddStats.Best, AddStats.Avg, AddStats.Worst);
			Console.WriteLine("Get()        {0,-6}     {1,-8}    {2,-8:F2}   {3,-8}", GetStats.Runs, GetStats.Best, GetStats.Avg, GetStats.Worst);
			

			Console.WriteLine("\r\n            Performance distribution");
			Console.WriteLine("            ------------------------");

			if(AddStats.Runs > 0)
				Console.WriteLine("Add()       {0,5:F1}% Best {1,5:F1}% Avg. {2,5:F1}% Worst -> ({3}, {4}, {5})", 
					new object[] {AddStats.PctBestCases, AddStats.PctAvgCases, AddStats.PctWorstCases, AddStats.BestCases, AddStats.AvgCases, AddStats.WorstCases });
			else
				Console.WriteLine("Add()        -");
			
			if(GetStats.Runs > 0)
				Console.WriteLine("Get()       {0,5:F1}% Best {1,5:F1}% Avg. {2,5:F1}% Worst -> ({3}, {4}, {5})",
					new object[] {	GetStats.PctBestCases, GetStats.PctAvgCases, GetStats.PctWorstCases, GetStats.BestCases, GetStats.AvgCases, GetStats.WorstCases });
			else
				Console.WriteLine("Get()        -");
			
			AddStats.Reset();
			GetStats.Reset();
		}

		/// <summary>
		/// Performing application specific clean-up, and disposal of allocated resources.
		/// </summary>
		private static void Dispose()
		{
			TxWriter.Close();
			StWriter.Close();
			FlStream.Close();
			_cache.Clear();
			_cache.Dispose();
		}		
	}
}

			

